#include "RectangleElement.h"
#include "Widget.h"

#include <algorithm>

Widget::Widget()
    : mLocation(DiscretePoint(0, 0)),
      mWidth(100),
      mHeight(100),
      mMaxHeight(0),
      mMyLayers(),
      mLayers(),
      mChildren(),
      mBackgroundColor(Color::Create(200, 0, 0))
{

}

Widget::~Widget()
{

}

std::unique_ptr<Widget> Widget::Create()
{
    return std::make_unique<Widget>();
}

void Widget::AddWidget(WidgetUPtr widget)
{
    mChildren.push_back(std::move(widget));
}

unsigned Widget::GetWidth() const
{
    return mWidth;
}

unsigned Widget::GetHeight() const
{
    return mHeight;
}

DiscretePoint Widget::GetLocation() const
{
    return mLocation;
}

std::vector<VisualLayer*> Widget::GetLayers() const
{
    return mLayers;
}

unsigned Widget::GetMaxHeight() const
{
    return mMaxHeight;
}

void Widget::SetMaxHeight(unsigned maxHeight)
{
    mMaxHeight = maxHeight;
}

void Widget::CopyMyLayers()
{
    mLayers.clear();
    mLayers.reserve(mMyLayers.size());
    auto getRaw = [](const VisualLayerUPtr& uptr){return uptr.get();};
    std::transform(mMyLayers.begin(), mMyLayers.end(), std::back_inserter(mLayers), getRaw);
}

void Widget::AddChildLayers(const PaintEvent* event)
{
    for(auto& child: mChildren)
    {
        child->SetSize(mWidth, mHeight);
        child->SetLocation(mLocation);
        child->OnPaintEvent(event);
        const auto layers = child->GetLayers();
        mLayers.insert(mLayers.end(), layers.begin(), layers.end());
    }
}

void Widget::OnPaintEvent(const PaintEvent* event)
{
    mMyLayers.clear();
    AddBackground(event);
    CopyMyLayers();
    AddChildLayers(event);
}

bool Widget::Contains(const DiscretePoint& loc) const
{
    if(loc.GetX() < mLocation.GetX()) return false;
    if(loc.GetX() > mLocation.GetX() + mWidth) return false;
    if(loc.GetY() > mLocation.GetY() + mHeight) return false;
    if(loc.GetY() < mLocation.GetY()) return false;
    return true;
}

bool Widget::OnKeyboardEvent(const KeyboardEvent* event)
{
    bool inChild = false;
    for(const auto& child : mChildren)
    {
        if(child->OnKeyboardEvent(event))
        {
            inChild = true;
            break;
        }
    }
    if(!inChild)
    {
        return OnMyKeyboardEvent(event);
    }
    return true;
}

bool Widget::OnMyKeyboardEvent(const KeyboardEvent* event)
{
    return false;
}

bool Widget::OnMouseEvent(const MouseEvent* event)
{
    bool inChild = false;
    for(const auto& child : mChildren)
    {
        if(child->OnMouseEvent(event))
        {
            inChild = true;
            break;
        }
    }
    if(!inChild)
    {
        if(Contains(event->GetClientLocation()))
        {
            OnMyMouseEvent(event);
            return true;
        }
        else
        {
            return false;
        }
    }
    return true;
}

void Widget::OnMyMouseEvent(const MouseEvent* event)
{

}

void Widget::AddBackground(const PaintEvent* event)
{
    auto shape = RectangleElement::Create(mLocation, mWidth, mHeight);
    shape->SetFillColor(Color::Create(*mBackgroundColor));
    auto shapeLayer = VisualLayer::Create();
    shapeLayer->SetShape(std::move(shape));
    mMyLayers.push_back(std::move(shapeLayer));
}

void Widget::SetBackgroundColor(ColorUPtr color)
{
    mBackgroundColor = std::move(color);
}

void Widget::SetLocation(const DiscretePoint& loc)
{
    mLocation = loc;
}

void Widget::SetSize(unsigned width, unsigned height)
{
    mWidth = width;
    mHeight = height;
}
